// Copyright Vladimir Prus 2002. Permission to copy, use,
// modify, sell and distribute this software is granted provided this
// copyright notice appears in all copies. This software is provided
// "as is" without express or implied warranty, and with no claim as
// to its suitability for any purpose.

#ifndef CMDLINE_VP_2003_05_19
#define CMDLINE_VP_2003_05_19

#include <boost/program_options/errors.hpp>

#include <boost/function.hpp>

#include <string>
#include <vector>

namespace boost { namespace program_options {

    class invalid_command_line_syntax : public invalid_syntax {
    public:
        enum kind_t {
            long_not_allowed,
            long_adjacent_not_allowed,
            short_adjacent_not_allowed,
            empty_adjacent_parameter,
            missing_parameter,
            extra_parameter,
        };

        invalid_command_line_syntax(const std::string& tokens, kind_t kind);
        kind_t kind() const;
    protected:
        static std::string error_message(kind_t kind);
    private:
        kind_t m_kind;        
    };

    class invalid_command_line_style : public error {
    public:
        invalid_command_line_style(const std::string& msg)
        : error(msg)
        {}
    };

    /** Command line parser class. Main requirements were:
        - Powerfull enough to support all common uses.
        - Simple and easy to learn/use.
        - Minimal code size and extrernal dependencies.
        - Extensible for custom syntaxes.

        First all options are registered. After that, elements of command line
        are extracted using operator++. 

        For each element, user can find
        - if it's an option or an argument
        - name of the option
        - index of the option
        - option value(s), if any
        
        Sometimes the registered option name is not equal to the encountered
        one, for example, because name abbreviation is supported.  Therefore
        two option names can be obtained: 
        - the registered one 
        - the one found at the command line

        There are lot of style options, which can be used to tune the command
        line parsing. In addition, it's possible to install additional parser
        which will process custom option styles.

        @todo mininal match length for guessing?
    */
    class cmdline {
    public:
        /** Various possible styles of options.

            There are "long" options, which start with "--" and "short",
            which start with either "-" or "/". Both kinds can be allowed or
            disallowed, see allow_long and allow_short. The allowed character
            for short option is also configurable.

            Option's value can be specified in the same token as value
            ("--foo=bar"), or in the next token.

            It's possible to introduce long option by the same character as
            long option, see allow_long_disguise.

            Finally, guessing (specifying only prefix of option) and case
            insensitive processing are supported.
        */
        enum style_t {
            /// Allow "--long_name" style
            allow_long = 1,
            /// Alow "-<single character" style
            allow_short = allow_long << 1,
            /// Allow "-" in short options
            allow_dash_for_short = allow_short << 1,
            /// Allow "/" in short options
            allow_slash_for_short = allow_dash_for_short << 1,
            /** Allow option parameter in the same token
                for long option, like in
                @verbatim
                --foo=10
                @endverbatim
            */
            long_allow_adjacent = allow_slash_for_short << 1,
            /** Allow option parameter in the same token for
               long options. */
            long_allow_next = long_allow_adjacent << 1,
            /** Allow option parameter in the same token for
               short options. */
            short_allow_adjacent = long_allow_next << 1,
            /** Allow option parameter in the next token for
                short options. */
            short_allow_next = short_allow_adjacent << 1,
            /** Allow to merge several short options together,
                so that "-s -k" become "-sk". All of the options
                but last should accept no parameter. For example, if
                "-s" accept a parameter, then "k" will be taken as
                parameter, not another short option. 
                Dos-style short options cannot be sticky.
            */
            allow_sticky = short_allow_next << 1,
            /** Allow abbreviated spellings for long options,
                if they unambiguously identify long option. 
                No long option name should be prefix of other 
                long option name is guessing is in effect.
            */
            allow_guessing = allow_sticky << 1,
            /** Ignore the difference in case for options. 
                @todo Should this apply to long options only?
             */            
            case_insentitive = allow_guessing << 1,
            /** Allow long options with single option starting character,
                e.g <tt>-foo=10</tt>
             */
            allow_long_disguise = case_insentitive << 1,
            /** The more-or-less traditional unix style. */
            unix_style = (allow_short | short_allow_adjacent | short_allow_next
                          | allow_long | long_allow_adjacent | long_allow_next
                          | allow_sticky | allow_guessing 
                          | allow_dash_for_short),
            /** The default style. */
            default_style = unix_style
        };

        typedef function1<std::pair<std::string, std::string>, 
                          const std::string&> 
            additional_parser;
        
        /** Constructs a command line parser for (argc, argv) pair. Uses
            style options passed in 'style', which should be binary or'ed values
            of style_t enum. It can also be zero, in which case a "default"
            style will be used. If 'allow_unregistered' is true, then allows 
            unregistered options. They will be assigned index 1 and are
            assumed to have optional parameter.
        */
        cmdline(const std::vector<std::string>& args, int style,
                bool allow_unregistered = false);

        /** @overload */
        cmdline(int argc, const char*const * argv, int style, 
                bool allow_unregistered = false);

        /** Set additional parser. This will be called for each token
            of command line. If first string in pair is not empty,
            then the token is considered matched by this parser,
            and the first string will be considered an option name
            (which can be long or short), while the second will be
            option's parameter (if not empty). 
            Note that additional parser can match only one token.
        */
        void set_additional_parser(additional_parser p);

        /** Registers a new option.
            @param long_name Long name to use. When ending '*', symbols up to
            it give an allowed prefix -- all options starting with it will be
            allowed. The first character may not be '-'. Empty string means no
            long name.
            @param short_name Short name to use. Value of '\0' means no short
            name.
            @param properties Tell about possible parameters
               '|' -- no parameter
               '?' -- optional parameter
               ':' -- required parameter
               '*' -- 0 or more parameters
               '+' -- 1 or more parameters
            @param index A distinguishing value for the option.
            The index will be returned by 'next' member function. Indices
            need not be unqiue -- e.g. client can set all indices to 0, and
            use string value to recognize options.
        */
        void add_option(const std::string& long_name, char short_name,
                        char properties = '|', int index = 0);

        /** @overload */
        void add_option(const char* long_name, char short_name,
                        char properties = '|', int index = 0);


        // Returns false when nothing more can be extracted
        operator bool() const;

        // Advances to the next element on command line. 
        // When called for the first time, positions on the first element.
        cmdline& operator++();

        /// Tells if the current element is option.
        bool at_option() const;

        /// Tells if the current element is argument.
        bool at_argument() const;

        /** Returns the option name. If there's long option name associated with
            this option, it is returned, even if short name was used in command line.
            Otherwise, the short name given to 'add_option' is returned with '-' prepended. 
            For purposes of simplicity, '-' is used even when dos-style short option 
            was found.
        */
        const std::string& option_name() const;
        /** Returns the index for the current option. */
        int option_index() const;
        /** Returns the option name as found on the command line. Any symbols
            that introduce the option, or delimit its parameter will be
            stripped. This function allows to work with allowed prefixes, in
            which case 'option_name' will return the prefix specification, and
            full option name should be queried explicitly.
            */
        const std::string& raw_option_name() const;
        /** Returns the first of option values. If there's more than one option
            throws multiple_values. If there are no options, returns empty 
            string. */
        const std::string& option_value() const;
        /** Returns all option values. */
        const std::vector<std::string>& option_values() const;
        /** Returns the argument. */
        const std::string& argument() const;

        /** Returns all arguments read by this command line parser. */
        const std::vector<std::string>& arguments() const;

        /** Returns the token that was current by the time 'operator++' was
            invoked the last time. */
        const std::string& last() const;

    private:

        // Properties of an option.
        enum properties_t {
            no_parameter = 1,
            /// 0 or 1 parameter
            allow_parameter,
            /// exactly 1 parameter
            require_parameter,
            /// 0 or more parameters
            allow_parameters,
            /// 1 or more parameters 
            require_parameters,            
        };

        enum element_kind_t {
            ek_option = 0,
            ek_argument
        };

        // General error status.
        enum error_type_t {
            no_error = 0,
            unknown_option,
            ambiguous_option,
            invalid_syntax,
        };

        // Detailed error status.
        enum error_description_t {
            ed_success = 0,
            ed_long_not_allowed,
            ed_long_adjacent_not_allowed,
            ed_short_adjacent_not_allowed,
            ed_empty_adjacent_parameter,
            ed_missing_parameter,
            ed_extra_parameter,
            ed_unknown_option,
            ed_ambiguous_option,
        };

        struct option {
            option(const std::string& long_name, char short_name,
                   properties_t properties, int index)
            : long_name(long_name), short_name(short_name), index(index),
            properties(properties)
            {}

            std::string long_name;
            char short_name;
            int index;
            properties_t properties;
        };
        std::vector<option> options;

        void init(const std::vector<std::string>& args, int style,
                  bool allow_unregistered);
        void check_style(int style) const;
       
        void next();

        const option* find_long_option(const char* name);
        const option* find_short_option(char name);

        enum option_kind { error_option, no_option, long_option, short_option,
                           dos_option };
        option_kind is_option(const char* s);
        // All handle_* member functions take string without any "--" or "-" or "/"
        // They return true and success and set m_num_tokens to the number of
        // tokens that were consumed.
        bool handle_long_option(const char* s);
        bool handle_short_option(const char* s, bool ignore_sticky = false);
        bool handle_dos_option(const char* s);
        // Attempts to apply additional parser to 's'.
        bool handle_additional_parser(const std::pair<std::string, std::string>& p);

        bool process_parameter(const option* opt, bool adjacent_parameter,
                              bool next_parameter);
        void advance(int count);

        /// Converts parameter property character into enum value.
        properties_t translate_property(char p);

        /** Clears all error state. If there were an error, throws appripriate
            exception. */
        void clear_error();

        // Copies of input.
        std::vector<std::string> args;
        style_t style;
        bool allow_unregistered;

        // Current state of parsing.
        unsigned index;
        const char* m_current;
        const char* m_next;
        const char* pending_short_option;
        bool m_no_more_options;
        error_description_t m_error_description;
        element_kind_t m_element_kind;
        int m_option_index;

        // Results of parsing the last option
        std::string m_last;
        const option* m_opt;
        std::string  m_option_name, m_raw_option_name, m_argument;
        std::vector<std::string> m_option_values; 
        int m_num_tokens;
        bool m_disguised_long;

        std::vector<std::string> m_arguments;

        additional_parser m_additional_parser;
    };
    
    namespace detail {
        void test_cmdline_detail();
    }

}}

#endif

