// Copyright Vladimir Prus 2002. Permission to copy, use,
// modify, sell and distribute this software is granted provided this
// copyright notice appears in all copies. This software is provided
// "as is" without express or implied warranty, and with no claim as
// to its suitability for any purpose.

#ifndef CONFIG_FILE_VP_2003_01_02
#define CONFIG_FILE_VP_2003_01_02

#include <iosfwd>
#include <string>
#include <set>

#include <boost/noncopyable.hpp>

namespace boost { namespace program_options {

    /** Standalone parser for config files in ini-line format:

        Config file can not contain arguments.


        - '#' is comment character: it is ignored together with
          the rest of the line.
        - variable assignments are in the form
          name '=' value.
          spaces around '=' are trimmed.
        - Section names are given in brackets. 

         The actuall variable name is constructed by combining current section
         name and variable name, with dot between. If section_name already
         contains dot at the end, new dot is not inserted. For example:
         @verbatim
         [gui.accessibility]
         visual_bell=yes
         @endverbatim
         will result in option "gui.accessibility.visual_bell" with value
         "yes" been returned.
     */    
    class config_file : noncopyable {
    public:

        /** Creates a config file parser for the specified stream.
            If 'validate' is set to false, allows all options. */
        config_file(std::istream& is, bool validate = true);
        /** Creates a config file parser for the specified file. */
        config_file(const std::string& filename, bool validate = true);

        ~config_file();

        /** Adds another allowed option. If the 'name' ends with
            '*', then all options with the same prefix are
            allowed. For example, if 'name' is 'foo*', then 'foo1' and
            'foo_bar' are allowed. */
        void add_option(const char* name);

        /** Returns true if there's another option to extract, i.e. if 
            name() and value() are valid. Is false immediately after 
            construction -- call operator++ to obtain the first value. */
        operator bool() const;

        /** Moves to the next option assignment in this file. When called
            for the first time, moves to the first assignment. 
            @throws invalid_syntax
            Can be called after exception is thrown, and will return
            next variable.
        */
        config_file& operator++();

        /** Returns the option name for the current assignment. */
        const std::string& name() const;

        /** Return the value for the current assignment. */
        const std::string& value() const;
    private:
        // Returns true if 's' is a registered option name.
        bool allowed_option(const std::string& s) const; 

        std::istream* is;
        // If 'own_istream', then 'is' was created in ctor and should
        // be deleted in dtor.
        bool own_istream;
        bool m_validate;
        std::set<std::string> allowed_options;
        // Invariant: no element is prefix of other element.
        std::set<std::string> allowed_prefixes;
        std::string m_name, m_value;
        std::string m_prefix;
    };


}}

#endif
