// Copyright Vladimir Prus 2002. Permission to copy, use,
// modify, sell and distribute this software is granted provided this
// copyright notice appears in all copies. This software is provided
// "as is" without express or implied warranty, and with no claim as
// to its suitability for any purpose.

#ifndef OPTION_DESCRIPTION_VP_2003_05_19
#define OPTION_DESCRIPTION_VP_2003_05_19

#include <boost/function.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/detail/workaround.hpp>
#include <boost/any.hpp>

#include <string>
#include <vector>
#include <set>
#include <map>
#include <stdexcept>

#include <iosfwd>

#include <boost/program_options/errors.hpp>

/** Boost namespace */
namespace boost { 
/** Namespace for the library. */
namespace program_options {


    /** Class thrown when value of option is incorrect. */
    class validation_error : public error {
    public:
        validation_error(const std::string& what) : error(what) {}
    };

    /** Describes one possible command line/config file option. There are two
        kinds of properties of an option. First describe it syntactically and
        are used only to validate input. Second affect interpretation of the
        option, for example default value for it or function that should be
        called  when the value is finally know. Routines which perform parsing
        never use second kind of properties -- they are side effect free.
        @sa options_description
    */
    class option_description {
    public:

        /** The type of function which perform validation of option parameters.
            It will be called with the current value of
            option as the first parameter (possible empty 'any'),
            and with values specified in the next occurence of
            option. It should update the value, or throw.
         */
        typedef boost::function2<void, boost::any&, 
                                 const std::vector<std::string>&> 
             validator_function;

        /** The type of function which will be called to notify about
            final value of option. */
        typedef boost::function1<void,  const boost::any&> notify_function;

        /** The type of function that is used to initialize option_description.
         */
        typedef boost::function1<void, option_description&> initialize_function;


        /** @name Construction */
        //@{
        option_description();

        /** Creates an option_description instance with
            the specified name and parameter_name. Name is
            set by call to this->name.
            Flags will be set this way:
            - '|' (no parameter) if parameter_name is empty
            - '?' (optional), '+' (1 or more values)
              '*' (0 or more values) if the last symbol
              of parameter_name is '?', '+', or '*'.
            - ':' (required parameter) otherwise.
        */
        option_description(const char* name,
                           const char* parameter_name);

        /** Same as above, but also sets the description
            for this option. */
        option_description(const char* name,
                           const char* parameter,
                           const char* description);

        /** Initializes *this from name, assuming
            empty parameter_name. After that,
            calls 'f' on *this. 
            @anchor initialize_func_ctor
        */
        option_description(const char* name,
                           const initialize_function& f);

        /** Initialized *this from name and description, and 
            assuming empty parameter_name. After that,
            calls 'f' on *this. */
        option_description(const char* name,
                           const initialize_function& f,
                           const char* description);

        virtual ~option_description() {}

        /** Sets the long name and short name, using the following
            convention:
            - if there's no "," character in 'name', it specifies long name
            - otherwise, the part before "," specifies long name and the part
            after -- long name.
         */
        option_description& name(const char* name);
        /// Sets the description
        option_description& description(const char* desc);
        /** Sets the parameter name
            @sa boost::program_options::parameter
        */
        option_description& parameter(const char* parameter);
        /** Sets flags to the passed value. The value of 'c' has the same
            meaning as in cmdline::add_option. */
        option_description& flags(char f);
        /// Sets default value of the option.
        option_description& default_value(const char*);
        /// Sets default value of the option's parameter.
        option_description& default_parameter(const char*);
        /** Sets validator to be used for the options's parameter.
        */
        option_description& validator(const validator_function&);
        /// Sets notify function to be used with this option.
        option_description& notify(const notify_function&);

        //@}

        /// @name Syntantic properties
        //@{

        /// Name to be used with short-style option ("-w").
        const std::string& short_name() const;

        /// Name to be used with long-style option ("--whatever").
        const std::string& long_name() const;
        /// Explanation of this option
        const std::string& description() const;
        /// Name of the parameter
        const std::string& parameter() const;
        /// Flags that apply to this variable.
        char flags() const;

        //@}

        /// @name Semantic properties
        //@{
        /// Returns default value
        const std::string& default_value() const;
        /// Returns detaul parameter's value
        const std::string& default_parameter() const;

        /// Function that performs validation and fixup on value.
        const validator_function& validator() const;

        /// Function that should be called when value of the variable is finally set.
        const notify_function& notify() const;

        //@}

        /// Returns the option name, formatted suitably for usage message. 
        std::string format_name() const;

        /** Return the parameter name and properties, formatted suitably for
            usage message. */
        std::string format_parameter() const;

    private:
        void init(const std::string& name, char param_flags, 
                  const std::string& param, const std::string& description);

        char guess_flags(const char* parameter);

    private:
        std::string m_short_name, m_long_name, m_description, m_parameter;
        char m_flags;
        std::string m_default_value, m_default_parameter;
        validator_function m_validator;
        notify_function m_notify;
    };

    class options_description;

    /** Class which provides convenient creation syntax to option_description.
        Any option description class which desires to have easy init syntax,
        should derive from this one. */
    template<typename Derived>
    class option_description_easy_init : public option_description {
    public:
        typedef Derived return_type;
        
        return_type& operator()(const char* name,
                               const char* parameter_name);
        
        return_type& operator()(const char* name,
                               const char* parameter_name,
                               const char* description);
            
        return_type& operator()(const char* name,
                               const option_description::initialize_function& f);
        
        return_type& operator()(const char* name,
                               const option_description::initialize_function& f,
                               const char* description);
       
        return_type& long_name(const char* name);
        return_type& short_name(const char* name);
        return_type& description(const char* desc);
        return_type& parameter(const char* parameter);
        return_type& flags(char f);
        return_type& default_value(const char* v);
        return_type& default_parameter(const char* p);
        return_type& validator(const validator_function& vf);
        return_type& notify(const notify_function& nf);

        void set_owner(options_description* owner);
    private:
        options_description* owner;
    };


    /** A set of option descriptions. This provides convenient interface for
        adding new option (the add_options) method, and facilities to search
        for options by name.
        
        See @ref a_adding_options "here" for option adding interface discussion.
        @sa option_description
    */
    class options_description {
    public:

        /// User-provided function for formatting option_description instances.
        typedef boost::function2<void, std::ostream&, const option_description&>
            formatter;

        /** Creates the instance. */
        options_description();
        /** Creates the instance. The 'caption' parameter gives the name of
            this 'options_description' instance. Primarily usefull for output.
        */
        options_description(const std::string& caption);
        /** Adds new variable description. Throws duplicate_variable_error if
            either short or long name matches that of already present one. 
        */
        void add(shared_ptr<option_description> desc);
        /** Adds a group of option description. This have the same
            effect as adding all option_descriptions in 'desc' 
            individually, except that output operator will show
            a separate group.
        */
        void add(const options_description& desc);

    private:
        class p_option_description : 
            public option_description_easy_init<p_option_description> {
        };

    public:
        /** Returns an object of implementation-defined type suitable for adding
            options to options_description. The object returned will have these
            methods:
            - call operator, with signatures matching those of option_description
              constructors. Effects: new option_description object is created and
              added.
            - flags, default_value, default_parameter, validator, notify. They
            signatures match those of methods in construction part of
            option_description's interface. Precondition: at least one object
            should be already added via current adder object. Effects: forwards
            call to the appropriate method of last added option_description.
        */
        option_description_easy_init<p_option_description> add_options();

#if ! BOOST_WORKAROUND(__BORLANDC__, BOOST_TESTED_AT(0x551)) && \
    ! BOOST_WORKAROUND(BOOST_INTEL_CXX_VERSION, BOOST_TESTED_AT(700))
        /** Returns an object suitable for adding options of user-defined class 
            T to options_description instance. This function is not available
            when using the borland's compiler.
        */
        template<class T>
        option_description_easy_init<T> add_options();
#endif

        /** Count the number of description with given name. Returns 0 or 1.
            The 'name' parameter can be either name of long option, and short
            option prepended by '-'.
        */
        unsigned count(const std::string& name) const;
        /** Count the number of descriptions having the given string as
            prefix. This makes sense only for long options.
        */
        unsigned count_approx(const std::string& prefix) const;
        /** Returns description given a name. 
            @pre count(name) == 1
        */
        const option_description& find(const std::string& name) const;
        /** Returns description given a prefix. Throws
            @pre count_approx(name) == 1 
        */
        const option_description& find_approx(const std::string& prefix) const;
        /// Returns all such strings x for which count(x) == 1
        std::set<std::string> keys() const;
        /** For each option description stored, contains long name if not empty,
            if it is empty, short name is returned.
        */
        std::set<std::string> primary_keys() const;
        /// Returns all such long options for which 'prefix' is prefix
        std::set<std::string> approximations(const std::string& prefix) const;

        /** Produces a human readable output of 'desc', listing options,
            their descriptions and allowed parameters. Other options_description
            instances previously passed to add will be output separately. */
        friend std::ostream& operator<<(std::ostream& os, 
                                        const options_description& desc);

        /** Output 'desc' to the specified streeam, calling 'f' to output each
            option_description element. */
        void output(std::ostream& os, formatter f) const;

    private:
        typedef std::map<std::string, int>::const_iterator name2index_iterator;
        typedef std::pair<name2index_iterator, name2index_iterator> 
            approximation_range;


        approximation_range find_approximation(const std::string& prefix) const;

        template<typename Derived>
        friend class option_description_easy_init;

        std::string m_caption;
        // Data organization is chosen since:
        // - there could be two names for one option
        // - option_add_proxy needs to know the last added option
        std::vector< shared_ptr<option_description> > options;
        std::map<std::string, int> name2index;
        // Whether the option comes from one of declared groups.
        std::vector<bool> belong_to_group;

        std::vector<options_description> groups;

    };

    /** Validates 's' and updates 'v'.
        @pre 'v' is either empty or in the state assigned by the previous
        invocation of 'validate'.
        Specializations are provided for bool, float, int, and string.
    */
    template<class T>
    class validator {
    public:
        /// Method that does the job.
        void operator()(boost::any& v, const std::vector<std::string>& s);
    };

    /** Validates sequences. Allows multiple values per option occurence
       and multiple occurences. */
    template<class T>
    class validator< std::vector<T> > {
    public:
        void operator()(boost::any& v, const std::vector<std::string>& s);
    };

    template<>
    void validator<bool>::operator()(any& v, const std::vector<std::string>& xs);

    template<>
    void validator<int>::operator()(any& v, const std::vector<std::string>& xs);

    template<>
    void validator<float>::operator()(any& v, 
                                      const std::vector<std::string>& xs);

    template<>
    void validator<std::string>::operator()(any& v, 
                                            const std::vector<std::string>& xs);

    /* Helpers for creating custom validators. */
    namespace validators {
        /* If v.size() > 1, throw validation_error. 
           If v.size() == 1, return v.front()
           Otherwise, returns a reference to a statically allocated
           empty string if 'allow_empty' and throws validation_error
           otherwise. */
        const std::string& get_single_string(const std::vector<std::string>& v, 
                                        bool allow_empty = false);

        /* Throws multiple_occurences if 'value' is not empty. */
        void check_first_occurence(const boost::any& value);
    }



    /* @defgroup g_options_parameter_creators Option parameter creators
        The purpose is those function is to ease creating of parameter
        description with validate input and optionally require result to
        be stored in some program variable. They all take either one or two
        parameters. First is the name of parameter; the same convention as for
        option_description ctor apply. It may not be empty.
        Second is address of a variable. It given notify field of option
        description will be set up to set the value of the variable.
        It is up to programmer to insure variable lifetime won't
        end before assigment.
        @todo Add assertions for non-empty name.
    */

    /** Creates initializer function. 
        For all types the function set validator to 'validator<T>'.
        For 'bool' additionaly sets default value to "0" and default
        parameter to "1", and also makes parameter optional.
        @sa @ref initialize_func_ctor "option_description constructor"
    */
    template<class T>
    option_description::initialize_function
    parameter(const char* s);

    /** Creates initializer function, which sets up validation for the
        type T, and established notify function which store the
        value in *v.
    */
    template<class T>
    option_description::initialize_function
    parameter(const char* s, T* v);

    /** Class thrown when duplicate option description is found. */
    class duplicate_option_error : public error {
    public:
        duplicate_option_error(const std::string& what) : error(what) {}
    };

    //CONSIDER: I'm not sure if such an exception is usefull. Maybe, assert?
    /* Class thrown when options is not found in options_description. */
    //class option_not_found_errror : public error {
    //public:
    //  option_not_found_error(const std::string& what) : error(what) {}
    //};

}}

#include "options_description.tcc"

#endif
