// Copyright Vladimir Prus 2002. Permission to copy, use,
// modify, sell and distribute this software is granted provided this
// copyright notice appears in all copies. This software is provided
// "as is" without express or implied warranty, and with no claim as
// to its suitability for any purpose.

#include <boost/lexical_cast.hpp>

#include <iomanip>

namespace boost { namespace program_options {

    template<class Derived>
    typename option_description_easy_init<Derived>::return_type&
    option_description_easy_init<Derived>::
    operator()(const char* name,
               const char* parameter_name)
    {
        // Use default ctor only, to minimized requirements on the derived type
        shared_ptr<option_description> d(new return_type);
        d->name(name);
        d->parameter(parameter_name);
        owner->add(d);
        static_cast<return_type&>(*d).set_owner(owner);
        return static_cast<return_type&>(*owner->options.back());
    }

    template<class Derived>
    typename option_description_easy_init<Derived>::return_type&
    option_description_easy_init<Derived>::
    operator()(const char* name,
               const char* parameter_name,
               const char* description)
    {
        shared_ptr<option_description> d(new return_type);
        d->name(name);
        d->parameter(parameter_name);
        d->description(description);        
        owner->add(d);
        static_cast<return_type&>(*d).set_owner(owner);
        return static_cast<return_type&>(*owner->options.back());
    }

    template<class Derived>
    typename option_description_easy_init<Derived>::return_type&
    option_description_easy_init<Derived>::
    operator()(const char* name,
               const option_description::initialize_function& f)
    {
        shared_ptr<option_description> d(new return_type);
        d->name(name);
        owner->add(d);
        static_cast<return_type&>(*d).set_owner(owner);
        f(*owner->options.back());
        return static_cast<return_type&>(*owner->options.back());
    }

    template<class Derived>
    typename option_description_easy_init<Derived>::return_type&
    option_description_easy_init<Derived>::
    operator()(const char* name,
               const option_description::initialize_function& f,
               const char* description)
    {
        shared_ptr<option_description> d(new return_type);
        d->name(name);
        d->description(description);
        owner->add(d);
        static_cast<return_type&>(*d).set_owner(owner);
        f(*owner->options.back());
        return static_cast<return_type&>(*owner->options.back());
    }

    template<class Derived>
    inline
    typename option_description_easy_init<Derived>::return_type& 
    option_description_easy_init<Derived>:: 
    long_name(const char* name)  
    { return static_cast<return_type&>(option_description::long_name(name)); }


    template<class Derived>
    inline
    typename option_description_easy_init<Derived>::return_type&
    option_description_easy_init<Derived>:: 
    short_name(const char* name)
    { return option_description::short_name(name); }

    template<class Derived>
    inline
    typename option_description_easy_init<Derived>::return_type&
    option_description_easy_init<Derived>:: 
    description(const char* desc)
    { return static_cast<return_type&>(option_description::description(desc)); }

    template<class Derived>
    inline
    typename option_description_easy_init<Derived>::return_type&
    option_description_easy_init<Derived>:: 
    parameter(const char* parameter)
    { return static_cast<return_type&>(option_description::parameter(parameter)); }

    template<class Derived>
    inline
    typename option_description_easy_init<Derived>::return_type&
    option_description_easy_init<Derived>:: 
    flags(char f)
    { return static_cast<return_type&>(option_description::flags(f)); }
        
    template<class Derived>
    inline
    typename option_description_easy_init<Derived>::return_type&
    option_description_easy_init<Derived>:: 
    default_value(const char* v)
    { return static_cast<return_type&>(option_description::default_value(v)); }
          
    template<class Derived>
    inline  
    typename option_description_easy_init<Derived>::return_type&
    option_description_easy_init<Derived>:: 
    default_parameter(const char* p)
    { return static_cast<return_type&>(option_description::default_parameter(p)); }

    template<class Derived>
    inline
    typename option_description_easy_init<Derived>::return_type&
    option_description_easy_init<Derived>:: 
    validator(const validator_function& vf)
    { return static_cast<return_type&>(option_description::validator(vf)); }

    template<class Derived>
    inline
    typename option_description_easy_init<Derived>::return_type&
    option_description_easy_init<Derived>:: 
    notify(const notify_function& nf)
    { return static_cast<return_type&>(option_description::notify_function(nf)); }

    template<class Derived>
    void
    option_description_easy_init<Derived>::set_owner(options_description* owner)
    {
        this->owner = owner;
    }

#if ! BOOST_WORKAROUND(__BORLANDC__, BOOST_TESTED_AT(0x551)) && \
    ! BOOST_WORKAROUND(BOOST_INTEL_CXX_VERSION, BOOST_TESTED_AT(700))
    template<class T>
    option_description_easy_init<T> 
    options_description::add_options()
    {       
        option_description_easy_init<T> r;
        r.set_owner(this);
        return r;                        
    }
#endif


    // Code having to to with validation.
    template<class T>
    void
    validator<std::vector<T> >::operator()(boost::any& v, 
                                           const std::vector<std::string>& s)
    {
        if (v.empty()) {
            v = boost::any(std::vector<T>());
        }
        std::vector<T>* tv = boost::any_cast< std::vector<T> >(&v);
        assert(tv);
        for (size_t i = 0; i < s.size(); ++i)
        {
            try {
                tv->push_back(boost::lexical_cast<T>(s[i]));
            }
            catch(const bad_lexical_cast& e) {
                throw validation_error(std::string("value ").append(s[i]).
                                       append(" is invalid"));
            }
        }
    }

    namespace detail {

        template<class T>
        class assign_callback {
        public:
            assign_callback(T* p) : p(p) {}
            void operator()(const boost::any& v) {
                *p = boost::any_cast<T>(v);
            }
        private:
            T* p;
        };
    }

    template<class T>
    class parameter_creator {
    public:
        parameter_creator(const char* name)
        : name(name), p(0)
        {}

        parameter_creator(const char* name, T* p)
        : name(name), p(p)
        {}

        void operator()(option_description& d) const
        {
            d.parameter(name);
            d.validator(validator<T>());
            if (p)
                d.notify(detail::assign_callback<T>(p));            
        }
    private:
        const char* name;
        T* p;
    };

    // FIXME: should move this to CPP file, once I figure out why gcc fails
    // to link in this case.
    // Hmm... gcc 3.3 gives duplicate symbol error if it's here,
    // and gcc 3.2 does not link. Let's make it inline for now.
    template<>
    inline
    void parameter_creator<bool>::operator()(option_description& d) const
    {
        d.parameter(name);
        if (*name && name[strlen(name)-1] != '?')
            d.flags('?');
        d.validator(validator<bool>());
        if (p)
            d.notify(detail::assign_callback<bool>(p));            
        d.default_parameter("1");
        d.default_value("0");
    }

    template<class T>
    option_description::initialize_function
    parameter(const char* s)
    {
        return parameter_creator<T>(s);
    }

    template<class T>
    option_description::initialize_function
    parameter(const char* s, T* v)
    {
        return parameter_creator<T>(s, v);
    }
}}
