// Copyright Vladimir Prus 2002. Permission to copy, use,
// modify, sell and distribute this software is granted provided this
// copyright notice appears in all copies. This software is provided
// "as is" without express or implied warranty, and with no claim as
// to its suitability for any purpose.

#ifndef PARSERS_VP_2003_05_19
#define PARSERS_VP_2003_05_19

// CONSIDER: Get rid from this includes and use
// forward declaration?
#include <boost/program_options/options_description.hpp>

#include <boost/program_options/errors.hpp>

#include <iosfwd>
#include <vector>
#include <utility>

namespace boost { namespace program_options {

    class cmdline;
    class config_file;

    /** Holds the result of parsing command line or any other source. The option
        naming depends on the entity which created the class instances, but the
        usual convention is the same as for cmdline::option_name.
    */
    class options_and_arguments {
    public:

        // Default copy ctor/operator= are ok.
        
        typedef std::vector< std::pair<std::string, 
                                       std::vector< std::vector<std::string> > > >
        options_t;

        /** Options given, in the order of occurence. First element is the name
            stored according to name conventions above. Second element is the
            value specified for the option. Note that none of parse* function
            below perform any calls to validator or to notify routine.
            Side effect is that it's always possible to tell is an value was
            specified, since "--foobar=" syntax is not allowed.
        */
        const options_t& options() const;
        /** Arguments given, in the order of occurence. */
        const std::vector< std::string >& arguments() const;
        /** Returns the result of calling 'get_value'. */
        const std::string& operator[](const std::string& name) const;
        /** Returns value of option or empty string if the option is not
            present. Returning empty string if option not found is
            for convenience; note that the function is const. This function
            recognizes option names as given in data source. 
            @throw multiple_values multiple_occurences
        */
        const std::string& get_value(const std::string& name) const;
        /** Returns all values of the specified option. If option is not present,
            returns an empty vector. 
            @throw multiple_occurences
        */
        const std::vector<std::string>& get_values(const std::string& name) const;
        /** Returns all values of all occurences of the specified option.
            If the option is not present, returns an empty vector. */
        const std::vector< std::vector< std::string > >&
        get_all_values(const std::string& name) const;
                   
        /** Checks if option is present. */
        unsigned count(const std::string& name) const;

        /** Adds new option occurence. */
        void add_option(const std::string& name, 
                        const std::vector<std::string>& value);
        /** Adds new argument. */
        void add_argument(const std::string& argument);

        /** If there are several occurences of option 'name',
            merges values in all of them together. Otherwise,
            has no effect.
        */
        void merge_occurences(const std::string& name);
    private:
        typedef std::vector< std::vector< std::string > > option_data;
        std::vector< std::pair<std::string, option_data > > m_options;
        std::vector< std::string > m_arguments;
        std::map<std::string, int> option2index;
    };

    /** Perform validation and executes all semantic actions, i.e. applying
        default values and calling 'notify' functions.
        Returns the semantic values for all options.
        @pre 'perform_semantic_actions' is called previously.
        @todo return auto_ptr
    */
    const std::map<std::string, boost::any>
    perform_semantic_actions(options_and_arguments& oa, 
                             const options_description& desc);
    

    // First, performs semantic actions for 'oa'. 
    // Then, stores in 'm' all options that are defined in 'desc'. 
    // All the options must be either untyped or have string type.
    void store(options_and_arguments& oa, 
               std::map<std::string, std::string>& m,
               const options_description& desc);

    /** Parses the command line using the 'cmd' object.
        Essentilally a convenience funciton, to store data in
        options_and_arguments instance.
        @anchor parse_cmdline_func
    */
    options_and_arguments
    parse_command_line(cmdline& cmd);

    typedef function1<std::pair<std::string, std::string>, const std::string&> ext_parser;

    /** Parses a command line, gives as a sequences of tokens
        in args. The program name is not included.
        Syntax of command line is determined by the 'style' argument,
        see the cmdline for details.
    */
    options_and_arguments
    parse_command_line(const std::vector<std::string> args,
                       const options_description&,
                       int style = 0,
                       function1<std::pair<std::string, std::string>, const std::string&> ext
                       = ext_parser());

    /** Parses a command line, performing no syntax checks. 
        See the 'allow_unregistered' parameter to the cmdline
        class for details. */
    options_and_arguments
    parse_command_line(const std::vector<std::string> args,
                       int style = 0,
                       function1<std::pair<std::string, std::string>, const std::string&> ext
                       = ext_parser());

    /** @overload */
    options_and_arguments
    parse_command_line(int argc, const char* argv[],
                       const options_description&,
                       int style = 0,
                       function1<std::pair<std::string, std::string>, const std::string&> ext
                       = ext_parser());

    /** @overload */
    options_and_arguments
    parse_command_line(int argc, const char* argv[],
                       int style = 0,
                       function1<std::pair<std::string, std::string>, const std::string&> ext
                       = ext_parser());

    /** @overload 
        @rationale It's common to declare "main" as taking non-const argv. This
        function is needed so that parse_command_line on non-const argv works.
     */
    options_and_arguments
    parse_command_line(int argc, char* argv[],
                       const options_description&,
                       int style = 0,
                       function1<std::pair<std::string, std::string>, const std::string&> ext
                       = ext_parser());

    /** @overload */
    options_and_arguments
    parse_command_line(int argc, char* argv[],
                       int style = 0,
                       function1<std::pair<std::string, std::string>, const std::string&> ext
                       = ext_parser());


    /** Parses the passed config_file instance and returns the results.
        @anchor parse_config_file_func
    */
    options_and_arguments
    parse_config_file(config_file& cf);


    /** Parse a config file. 
    */
    options_and_arguments
    parse_config_file(std::istream&, const options_description&);

    /** @overload */
    options_and_arguments
    parse_config_file(std::istream&);
}}


#endif
