// Copyright Vladimir Prus 2002. Permission to copy, use,
// modify, sell and distribute this software is granted provided this
// copyright notice appears in all copies. This software is provided
// "as is" without express or implied warranty, and with no claim as
// to its suitability for any purpose.

#ifndef VARIABLES_MAP_VP_2003_05_19
#define VARIABLES_MAP_VP_2003_05_19

#include <boost/program_options/parsers.hpp>

#include <boost/lexical_cast.hpp>
#include <boost/any.hpp>

#include <string>
#include <map>

namespace boost { namespace program_options {

    /** Class holding value of option. Contains details about how the 
        value is set and allows to conveniently obtain the value.
    */
    class variable_value {
    public:
        variable_value() {}
        variable_value(const boost::any& v, bool defaulted) 
        : v(v), m_defaulted(defaulted) 
        {}

        /** If stored value if of type T, returns that value. Otherwise,
            throws boost::bad_any_cast exception. */
        template<class T> T as() const;

        /// Returns true if no value is stored.
        bool empty() const;
        /** Returns true if the value was not explcitly
            given, but has default value. */
        bool defaulted() const;
        /** Returns the contained value. */
        const boost::any& value() const;
    private:
        boost::any v;
        bool m_defaulted;
    };

    /** Implements string->string mapping with convenient value casting
        facilities. */
    class abstract_variables_map {
    public:
        abstract_variables_map();
        abstract_variables_map(const abstract_variables_map* next);

        virtual ~abstract_variables_map() {}

        /** Obtains the value of variable 'name', from *this and
            possibly from the chain of variable maps.

            - if there's no value in *this.
                - if there's next variable map, returns value from it
                - otherwise, returns empty value

            - if there's defaulted value
                - if there's next varaible map, which has non-defauled
                  value, return that
                - otherwise, return value from *this

            - if there's non-defauled value, returns it.
        */
        variable_value operator[](const std::string& name) const;

        /** Sets next variable map, which will be used to find
           variables not found in *this. */
        void next(abstract_variables_map* next);

    private:
        /** Returns value of variable 'name' stored in *this, or
            empty value otherwise. */
        virtual variable_value get(const std::string& name) const = 0;

        const abstract_variables_map* m_next;
    };

    /** Concrete variables map which store variables in real map. */
    class variables_map : public abstract_variables_map,
                          public std::map<std::string, variable_value>
    {
    public:
        variables_map();
        variables_map(const abstract_variables_map* next);

        // Resolve conflict between inherited operators.
        variable_value operator[](const std::string& name) const
        { return abstract_variables_map::operator[](name); }

    private:
        /** Implementation of abstract_variables_map::get
            which does 'find' in *this. */
        variable_value get(const std::string& name) const;
    };

    /** First, performs semantic actions for 'oa'. 
        Then, stores in 'm' all options that are defined in 'desc'. */
    void store(options_and_arguments& oa, variables_map& m,
               const options_description& desc);


    /*
     * Templates/inlines
     */

    inline bool
    variable_value::empty() const
    {
        return v.empty();
    }

    inline bool
    variable_value::defaulted() const
    {
        return m_defaulted;
    }

    inline
    const boost::any&
    variable_value::value() const
    {
        return v;
    }

    template<class T>
    T
    variable_value::as() const {
        return boost::any_cast<T>(v);
    }
}}


#endif
